import { Settings } from './models/avatarSettingsModel';
import {
    FileUploadSettingsFormView,
} from './views/fileUploadSettingsFormView';
import {
    ServiceSettingsFormView,
} from './views/avatarServiceSettingsFormView';
import { SettingsFormView } from './views/avatarSettingsFormView';


/* Define a namespace for Djblets.Avatars. */
export const Avatars = {
    FileUploadSettingsFormView,
    ServiceSettingsFormView,
    Settings,
    SettingsFormView,
};
