/**
 * Base class for a list used for config forms.
 */

import { BaseModel, spina } from '@beanbag/spina';


/**
 * Base class for a list used for config forms.
 */
@spina
export class List extends BaseModel {
}
