!function(){var e,t;e=this,t=function(e){"use strict";e.buildURL=function(){var e=0<arguments.length&&void 0!==arguments[0]?arguments[0]:{};let t=e.baseURL||"";var n=e.queryData;if(n){let e;"string"==typeof n?0===(e=n).indexOf("?")&&(e=e.substr(1)):e=$.param(n),e&&(t+="?"+e)}return(n=e.anchor)&&(0===n.indexOf("#")?t+=n:t+="#"+n),t},e.parseQueryString=function(e){var o=(1<arguments.length&&void 0!==arguments[1]?arguments[1]:{}).allowMultiValue;let r=e.indexOf("?"),t;t=-1===r?e:e.substr(r+1);var i={};if(0!==t.length){var s=t.split("&");for(let n=0;n<s.length;n++){var l=s[n];let e,t;r=l.indexOf("="),t=-1===r?(e=decodeURIComponent(l),null):(e=decodeURIComponent(l.substr(0,r)),decodeURIComponent(l.substr(r+1))),o&&i.hasOwnProperty(e)?_.isArray(i[e])?i[e].push(t):i[e]=[i[e],t]:i[e]=t}}return i}},"object"==typeof exports&&"undefined"!=typeof module?t(exports):"function"==typeof define&&define.amd?define(["exports"],t):t((e="undefined"!=typeof globalThis?globalThis:e||self).Djblets=e.Djblets||{}),Promise.withResolver=function(){let n;return[new Promise((e,t)=>{n=e}),n]}}.call(this);
