{% load djblets_forms djblets_utils %}

{% block pre_fields %}{% endblock %}

{{form.non_field_errors}}

{% for fieldset_title, fieldset in form|get_fieldsets %}
{%  include "djblets_forms/form_fieldset.html" with field_template="configforms/config_page_form_field.html" %}
{% empty %}
{%  for field in form %}
{%   include "configforms/config_page_form_field.html" %}
{%  endfor %}
{% endfor %}

{% block post_fields %}{% endblock %}

{% if form.save_label %}
<div class="djblets-c-config-forms-actions">
 <button class="ink-c-button" type="submit">
  {{form.save_label}}
 </button>
</div>
{% endif %}
